import md5, random
from cPickle import load, dump

DEBUG = 0
MAX_ITERATIONS = 99999

# The higher, the more we don't like being just net to water
WATER_FACTOR = 2
ROBOT_FACTOR = 1

def hdiff(a,b):
    # Use simple Manhattan distance
    return (abs(a[0] - b[0]) + abs(a[1] - b[1]))

class Problem:
    def __init__(self,start,goal,map,randness):
        """
        start, goal: start, goal position, in Y,X format
        map:  [y][x] array
        adj:  ditionary, (y,x), with penalty for entering that square
        """
        self.start, self.goal, self.map, self.randness = start,goal,map,randness
        stats.finds += 1

    def getCached(self):
        "Is this thing cached?"
        if self.randness: return None
        try:
            r = cache[self.start,self.goal]
            stats.hits += 1
            return r
        except KeyError:
            try:
                l = cache[self.goal,self.start][:]
                l.reverse()
                stats.reverseHits += 1
                return l
            except KeyError:
                pass
        return None

    def createStartNode(self, coord):
        return Node(coord, hdiff(coord, self.goal))

    DIFFS = (0,+1), (0,-1), (+1,0), (-1,0)

    TDIFFS = (0,+1), (0,-1), (+1,0), (-1,0), \
            (+1,+1), (-1,+1), (-1,+1), (-1,-1)

    def allNearbySquares(self, coord):
        r = ""
        for dx,dy in self.TDIFFS:
            r += self.map[coord[0] + dy][coord[1] + dx]
        return r

    def generateSuccessors(self, cur):
        l = []
        for dy, dx in self.DIFFS :
            y = cur.coord[0] + dy
            x = cur.coord[1] + dx
            if self.map[y][x] in '.@':
                node = Node((y,x), hdiff((y,x), self.goal), cur)
                # We don't like water
                nearby = self.allNearbySquares((y,x))
                node.f += nearby.count('~') * WATER_FACTOR
                node.f += nearby.count('*') * ROBOT_FACTOR

                # We don't like changing direction
                # (This algorithm doesn't work)
                if 0 and cur.parent:
                    node.f += abs(cur.parent.coord[0] - cur.coord[0]) != abs(dy)
                    node.f += abs(cur.parent.coord[1] - cur.coord[1]) != abs(dx)
                if self.randness: node.f += random.randrange(4)
                
                l.append(node)
        return l
            

class Node:
    def __init__(self,coord,h,parent=None):
        self.coord = coord
        self.parent = parent

        # Cost of getting to this node
        if self.parent:
            self.g = self.parent.g + 1
        else:
            self.g = 0

        self.f = self.g + h

    def makeList(self):
        "Go backwards from node to generate our coord list"
        l = []; node = self
        while node:
            l.append(node.coord)
            node = node.parent
        l.reverse()
        return l

    def __hash__(self):
        return hash(self.coord)

    def __eq__(self, other):
        return self.coord == other.coord

    def __str__(self):
        return '(%d,%d @%d/%d)' % (self.coord + (self.g, self.f))

    __repr__ = __str__

cache = {}
partials = {}                           # true if this was a partial cache element
class stats:
    hits = 0
    reverseHits = 0
    finds = 0
    partials = 0

def mapFile(map):
    m = md5.new()
    for x in map:
        m.update(x)
    return 'cache-%s' % m.hexdigest()

def storeCache(map):
    "Store cache to disk"
    dump(cache, open(mapFile(map), 'w'), 1)

def loadCache(map):
    global cache
    try:
        cache = load(open(mapFile(map)))
    except IOError:
        pass

def getStats():
    return 'A*: %d entries, %d+%d hits, %d misses / %d finds (%d%%)' % (
        len(cache), stats.hits, stats.reverseHits, (stats.finds - (stats.hits + stats.reverseHits)),
#        stats.partials,
        stats.finds,
        (100*(stats.hits + stats.reverseHits))/(max(1,stats.finds)))

# http://www.geocities.com/jheyesjones/pseudocode.html    
def astar(start, goal, map, randness = 0):

    problem = Problem(start, goal, map, randness)
    r = problem.getCached()
    if r:
        return r
    
    open = {}                           # By hash(node)
    openlist = []                       # Sorted by priority
    closed = {}                         # By hash(node)
    # 3 Put node_start on the open list
    node = problem.createStartNode(start)
    openlist.append(node)
    open[node] = node

    iterations = 0
                    
    # 4 while the OPEN list is not empty
    while openlist:
        if iterations > MAX_ITERATIONS:
            l = cur.makeList()
            if not randness: cache[start,cur.coord] = l
            return l
        iterations += 1
        
        # 6 Get the node off the open list with the lowest f and call it node_current
        cur = openlist.pop(0)
        del open[cur]
        
        # if node_current is the same state as node_goal we have found the solution; break from the while loop
        if cur.coord == goal:
            l = cur.makeList()
            if not randness: cache[start,goal] = l
            # (no) Cache all the partial paths as well
            #            for x in xrange(0,len(l)-1):
            #                cache[l[x], goal] = l[x:]
            #                if x > 0:
            #                    partials[l[x],goal] = 1
            return l

        # 9 for each node_successor of node_current
        if DEBUG: print "Sucessors of %s" % cur
        for succ in problem.generateSuccessors(cur):
            if DEBUG: print " %s" % succ
            # 12 find node_successor on the OPEN list
            # 13  if node_successor is on the OPEN list but the existing one is better then discard this successor and continue
            alreadyOpen = open.get(succ)
            if alreadyOpen and alreadyOpen.f <= succ.f:
                if DEBUG: print "  Already better pending (%d vs new %d)" % (alreadyOpen.f, succ.f)
                continue
            elif alreadyOpen:
                if DEBUG: print "  Pending but old.f=%d > f=%d" % (alreadyOpen.f, succ.f)
            # 14 if node_successor is on the CLOSED list but the existing one is better then discard this successor and continue
            if closed.has_key(succ) and closed[succ].f <= succ.f:
                if DEBUG: print "  Already better closed (%d vs new %d)" % (closed[succ].f, succ.f)
                continue

            # 15 Remove from OPEN and CLOSED and re-add
            if alreadyOpen:
                openlist.remove(alreadyOpen)
            try:
                del closed[succ]
            except KeyError:
                pass
            
            open[succ] = succ
            openlist.append(succ)

            if DEBUG: print "  not in list, added"

        if DEBUG: print "open: ", open.keys()
        closed[cur] = cur
        openlist.sort(lambda x,y: cmp(x.f,y.f) or cmp(id(x), id(y)))
    return []

maze = [
    #012345
    "######",  # 0
    "#....#",  # 1
    "###.##",  # 2
    "#....#",  # 3
    "#.#.##",  # 4
    "######"   # 5
    #012345
    ]                                   # maze[y][x]

GOAL = (1, 4)                           # YX
START = (4, 1)

def main():
    print astar(START, GOAL, maze)

if __name__ == '__main__':
    import util
    util.run(main)    



